from ya_market_api.feedback.const import ReactionStatus
from ya_market_api.base.dataclass import BaseResponse

from typing import Optional, Collection, FrozenSet, Dict, Any, List, ClassVar, overload
from warnings import warn

from pydantic.main import BaseModel
from pydantic.fields import Field
from pydantic.config import ConfigDict
from pydantic.functional_validators import field_validator
from pydantic.functional_serializers import field_serializer
from arrow import Arrow, get as get_arrow


class Request(BaseModel):
	QUERY_PARAMS: ClassVar[FrozenSet[str]] = frozenset({"limit", "page_token"})		# pydantic model_dump requires set
	S11N_ALIAS_FEEDBACK_IDS: ClassVar[str] = "feedbackIds"
	model_config = ConfigDict(arbitrary_types_allowed=True)

	# query params
	limit: Optional[int] = Field(default=None)
	page_token: Optional[str] = None

	# payload
	datetime_from: Optional[Arrow] = Field(default=None, serialization_alias="dateTimeFrom")
	datetime_to: Optional[Arrow] = Field(default=None, serialization_alias="dateTimeTo")
	paid: Optional[bool] = None
	rating_values: Optional[Collection[int]] = Field(default=None, serialization_alias="ratingValues")
	reaction_status: Optional[ReactionStatus] = Field(default=None, serialization_alias="reactionStatus")

	feedback_ids: Optional[Collection[int]] = Field(
		default=None,
		serialization_alias=S11N_ALIAS_FEEDBACK_IDS,
		min_length=1,
		max_length=50,
	)

	@overload
	def __init__(
		self,
		*,
		limit: Optional[int] = None,
		page_token: Optional[str] = None,
		feedback_ids: Optional[Collection[int]] = None,
	) -> None: ...
	@overload
	def __init__(
		self,
		*,
		limit: Optional[int] = None,
		page_token: Optional[str] = None,
		datetime_from: Optional[Arrow] = None,
		datetime_to: Optional[Arrow] = None,
		paid: Optional[bool] = None,
		rating_values: Optional[Collection[int]] = None,
		reaction_status: Optional[ReactionStatus] = None,
	) -> None: ...
	def __init__(
		self,
		*,
		limit: Optional[int] = None,
		page_token: Optional[str] = None,
		feedback_ids: Optional[Collection[int]] = None,
		datetime_from: Optional[Arrow] = None,
		datetime_to: Optional[Arrow] = None,
		paid: Optional[bool] = None,
		rating_values: Optional[Collection[int]] = None,
		reaction_status: Optional[ReactionStatus] = None,
	) -> None:
		super().__init__(
			limit=limit,
			page_token=page_token,
			feedback_ids=feedback_ids,
			datetime_from=datetime_from,
			datetime_to=datetime_to,
			paid=paid,
			rating_values=rating_values,
			reaction_status=reaction_status,
		)

	@field_validator("limit", mode="after")
	@classmethod
	def limit_is_valid(cls, value: Optional[int]) -> Optional[int]:
		if value is None:
			return None

		if value < 1 or value > 50:
			raise ValueError("The limit cannot be less than 1 or greater than 50")

		return value

	@field_validator("rating_values", mode="after")
	@classmethod
	def rating_values_are_valid(cls, value: Optional[Collection[int]]) -> Optional[Collection[int]]:
		if value is None:
			return value

		for i in value:
			if i < 1 or i > 5:
				raise ValueError(f"{i} is not valid rating value. Must be in range [1..5]")

		return value

	@field_serializer("datetime_from", "datetime_to", mode="plain")
	def serialize_optional_arrow(self, value: Optional[Arrow]) -> Optional[str]:
		if value is None:
			return None

		return value.isoformat()

	@field_serializer("reaction_status", mode="plain")
	def serialize_reaction_status(self, value: Optional[ReactionStatus]) -> Optional[str]:
		if value is None:
			return None

		return value.value

	def model_dump_request_params(self) -> Dict[str, Any]:
		return self.model_dump(include=self.QUERY_PARAMS, by_alias=True, exclude_none=True)

	def model_dump_request_payload(self) -> Dict[str, Any]:
		result = self.model_dump(exclude=self.QUERY_PARAMS, by_alias=True, exclude_none=True)

		if self.feedback_ids is not None and len(result) != 1:
			warn("When using feedback_id, the other parameters will be ignored")
			result = {self.S11N_ALIAS_FEEDBACK_IDS: result[self.S11N_ALIAS_FEEDBACK_IDS]}

		return result


class FeedbackIdentifiers(BaseModel):
	model_id: Optional[int] = Field(default=None, validation_alias="modelId", deprecated=True)
	order_id: Optional[int] = Field(default=None, validation_alias="orderId")


class FeedbackStatistics(BaseModel):
	comments_count: int = Field(validation_alias="commentsCount")
	rating: int
	paid_amount: Optional[int] = Field(default=None, validation_alias="paidAmount")
	recommended: Optional[bool] = None


class FeedbackDescription(BaseModel):
	advantages: Optional[str] = None
	comment: Optional[str] = None
	disadvantages: Optional[str] = None


class FeedbackMedia(BaseModel):
	photos: List[str] = Field(default_factory=list)
	videos: List[str] = Field(default_factory=list)


class Feedback(BaseModel):
	model_config = ConfigDict(arbitrary_types_allowed=True)

	id: int = Field(validation_alias="feedbackId")
	created_at: Arrow = Field(validation_alias="createdAt")
	identifiers: FeedbackIdentifiers
	need_reaction: bool = Field(validation_alias="needReaction")
	statistics: FeedbackStatistics
	author: Optional[str] = None
	description: Optional[FeedbackDescription] = None
	media: Optional[FeedbackMedia] = None

	@field_validator("created_at", mode="before")
	@classmethod
	def validate_arrow(cls, value: Any) -> Arrow:
		if not isinstance(value, str):
			raise ValueError("Raw value must be of str type")

		return get_arrow(value)


class Paging(BaseModel):
	next_page_token: Optional[str] = Field(default=None, validation_alias="nextPageToken")


class ResponseResult(BaseModel):
	feedbacks: List[Feedback]
	paging: Optional[Paging] = None


class Response(BaseResponse):
	result: Optional[ResponseResult] = None



from ya_market_api.guide.router import GuideRouter


class TestGuideRouter:
	def test_token_info(self):
		router = GuideRouter("")
		assert router.token_info() == "/auth/token"

	def test_delivery_services(self):
		router = GuideRouter("")
		assert router.delivery_services() == "/delivery/services"



from ya_market_api.base.router import Router


class OrderRouter(Router):
	def order_get(self, campaign_id: int, order_id: int) -> str:
		return f"{self.base_url}/campaigns/{campaign_id}/orders/{order_id}"



from ya_market_api.generic.functools import cache
from ya_market_api.base.router import Router


class GuideRegionRouter(Router):
	@cache
	def region_countries(self) -> str:
		return f"{self.base_url}/regions/countries"

	@cache
	def region_search(self) -> str:
		return f"{self.base_url}/regions"

	def region_info(self, region_id: int) -> str:
		return f"{self.base_url}/regions/{region_id}"

	def region_children(self, region_id: int) -> str:
		return f"{self.base_url}/regions/{region_id}/children"



from ya_market_api.order.base_api import BaseOrderAPI
from ya_market_api.base.config import Config
from ya_market_api.exception import CampaignIdError

import pytest


class TestBaseOrderAPI:
	def test_campaign_id(self):
		config = Config(None, "")
		api = BaseOrderAPI(config)

		with pytest.raises(CampaignIdError, match="The campaign_id was not specified"):
			api.campaign_id

		config.campaign_id = 512
		assert api.campaign_id == 512



from ya_market_api.feedback.router import FeedbackRouter


class TestFeedbackRouter:
	def test_feedback_list(self):
		router = FeedbackRouter("")
		assert router.feedback_list(512) == "/businesses/512/goods-feedback"

	def test_feedback_comment_list(self):
		router = FeedbackRouter("")
		assert router.feedback_comment_list(512) == "/businesses/512/goods-feedback/comments"

	def test_feedback_comment_add(self):
		router = FeedbackRouter("")
		assert router.feedback_comment_add(512) == "/businesses/512/goods-feedback/comments/update"

	def test_feedback_comment_update(self):
		router = FeedbackRouter("")
		assert router.feedback_comment_update(512) == "/businesses/512/goods-feedback/comments/update"

	def test_feedback_comment_delete(self):
		router = FeedbackRouter("")
		assert router.feedback_comment_delete(512) == "/businesses/512/goods-feedback/comments/delete"

	def test_feedback_reaction_skip(self):
		router = FeedbackRouter("")
		assert router.feedback_reaction_skip(512) == "/businesses/512/goods-feedback/skip-reaction"



from ya_market_api.base.async_api_mixin import AsyncAPIMixin
from ya_market_api.feedback.base_api import BaseFeedbackAPI
from ya_market_api.feedback.dataclass import (
	FeedbackListRequest, FeedbackListResponse, FeedbackCommentListRequest, FeedbackCommentListResponse,
	FeedbackCommentAddRequest, FeedbackCommentAddResponse, FeedbackCommentUpdateRequest, FeedbackCommentUpdateResponse,
	FeedbackCommentDeleteRequest, FeedbackCommentDeleteResponse, FeedbackReactionSkipRequest,
	FeedbackReactionSkipResponse,
)

from typing import Optional


class AsyncFeedbackAPI(AsyncAPIMixin, BaseFeedbackAPI):
	async def get_feedback_list(self, request: Optional[FeedbackListRequest] = None) -> FeedbackListResponse:
		request = request or FeedbackListRequest()
		url = self.router.feedback_list(self.business_id)

		async with self.session.post(
			url=url,
			params=request.model_dump_request_params(),
			json=request.model_dump_request_payload(),
		) as response:
			self.validate_response(response)
			return FeedbackListResponse.model_validate_json(await response.text())

	async def get_feedback_comment_list(self, request: FeedbackCommentListRequest) -> FeedbackCommentListResponse:
		url = self.router.feedback_comment_list(self.business_id)

		async with self.session.post(
			url=url,
			params=request.model_dump_request_params(),
			json=request.model_dump_request_payload(),
		) as response:
			self.validate_response(response)
			return FeedbackCommentListResponse.model_validate_json(await response.text())

	async def add_feedback_comment(self, request: FeedbackCommentAddRequest) -> FeedbackCommentAddResponse:
		url = self.router.feedback_comment_add(self.business_id)

		async with self.session.post(
			url=url,
			json=request.model_dump(by_alias=True, exclude_defaults=True),
		) as response:
			self.validate_response(response)
			return FeedbackCommentAddResponse.model_validate_json(await response.text())

	async def update_feedback_comment(self, request: FeedbackCommentUpdateRequest) -> FeedbackCommentUpdateResponse:
		url = self.router.feedback_comment_update(self.business_id)

		async with self.session.post(
			url=url,
			json=request.model_dump(by_alias=True, exclude_defaults=True),
		) as response:
			self.validate_response(response)
			return FeedbackCommentUpdateResponse.model_validate_json(await response.text())

	async def delete_feedback_comment(self, request: FeedbackCommentDeleteRequest) -> FeedbackCommentDeleteResponse:
		url = self.router.feedback_comment_delete(self.business_id)

		async with self.session.post(url, json=request.model_dump(by_alias=True, exclude_defaults=True)) as response:
			self.validate_response(response)
			return FeedbackCommentDeleteResponse.model_validate_json(await response.text())

	async def skip_feedback_reaction(self, request: FeedbackReactionSkipRequest) -> FeedbackReactionSkipResponse:
		url = self.router.feedback_reaction_skip(self.business_id)

		async with self.session.post(url, json=request.model_dump(by_alias=True, exclude_defaults=True)) as response:
			self.validate_response(response)
			return FeedbackReactionSkipResponse.model_validate_json(await response.text())



from ya_market_api.base.router import Router
from ya_market_api.base.config import Config

from abc import ABC, abstractmethod
from typing import TypeVar, Generic


RouterT = TypeVar("RouterT", bound=Router)


class API(ABC, Generic[RouterT]):
	config: Config
	router: RouterT

	def __init__(self, config: Config) -> None:
		self.config = config
		self.router = self.make_router(config.base_url)

	@staticmethod
	@abstractmethod
	def make_router(base_url: str) -> RouterT: ...



from ya_market_api.base.enum_toolkit import allow_unknown

from enum import Enum


@allow_unknown
class ReactionStatus(Enum):
	ALL = "ALL"
	NEED_REACTION = "NEED_REACTION"


@allow_unknown
class CommentAuthorType(Enum):
	USER = "USER"
	BUSINESS = "BUSINESS"


@allow_unknown
class CommentStatus(Enum):
	PUBLISHED = "PUBLISHED"
	UNMODERATED = "UNMODERATED"
	BANNED = "BANNED"
	DELETED = "DELETED"



from contextlib import AbstractAsyncContextManager
from types import TracebackType
from typing import Type, Optional, Union, Literal


class FakeResponse(AbstractAsyncContextManager):
	def __init__(self, text: str) -> None:
		self._text = text

	async def __aenter__(self):
		return self

	async def __aexit__(
		self,
		exc_type: Optional[Type[BaseException]],
		exc_value: Optional[BaseException],
		traceback: Optional[TracebackType],
	) -> None:
		return None

	async def text(self) -> str:
		return self._text


class FakeAsyncSession:
	last_call_method: Optional[Literal["GET", "POST"]]
	last_call_url: Optional[str]
	last_call_json: Union[None, str, dict]
	last_call_params: Optional[dict]

	def __init__(self, response_text: str) -> None:
		self.response = FakeResponse(response_text)
		self.last_call_method = None
		self.last_call_url = None
		self.last_call_params = None
		self.last_call_json = None

	def post(self, url: str, json: Union[str, dict], params: Optional[dict] = None) -> FakeResponse:
		self.last_call_method = "POST"
		self.last_call_url = url
		self.last_call_json = json
		self.last_call_params = params
		return self.response

	def get(self, url: str, params: Optional[dict] = None) -> FakeResponse:
		self.last_call_method = "GET"
		self.last_call_url = url
		self.last_call_params = params
		return self.response



from enum import Enum
from typing import Type, Any, TypeVar


T = TypeVar("T", bound=Type[Enum])


def allow_unknown(cls: T) -> T:
	"""
	Extends existing enumeration, which allows it to interpret all unspecified values as UNKNOWN.

	Examples:
		>>> from enum import Enum
		>>> @allow_unknown
		>>> class A(Enum):
		>>> 	A = "A"
		>>> print(A("A"))		# "A.A"
		>>> print(A("B"))		# "A.UNKNOWN"
		>>> print(A.UNKNOWN)		# "A.UNKNOWN"
		>>> print(A.A)		# "A.UNKNOWN"

	Notes:
		* Static code checkers don't know about the "UNKNOWN" option and won't let you use it.
	"""
	values = {"UNKNOWN": "UNKNOWN"}

	for i in cls:
		values[i.name] = i.value


	class MissingProxy(Enum):
		@classmethod
		def _missing_(cls, value: Any):
			return cls.UNKNOWN


	return Enum(cls.__name__, values, type=MissingProxy)



from ya_market_api.exception import InvalidResponseError, AuthorizationError, NotFoundError
from ya_market_api.base.async_config import AsyncConfig

from http import HTTPStatus

from aiohttp.client import ClientSession, ClientResponse


class AsyncAPIMixin:
	config: AsyncConfig

	def __init__(self, config: AsyncConfig) -> None:
		super().__init__(config)

	@property
	def session(self) -> ClientSession:
		return self.config.session

	def validate_response(self, response: ClientResponse) -> None:
		if not response.ok:
			if response.status == HTTPStatus.FORBIDDEN or response.status == HTTPStatus.UNAUTHORIZED:
				raise AuthorizationError("Unauthorized")
			elif response.status == HTTPStatus.NOT_FOUND:
				raise NotFoundError("Resource was not found")

			raise InvalidResponseError("Response is not valid")



from ya_market_api.offer.sync_api import SyncOfferAPI
from ya_market_api.offer.dataclass import OfferListByBusinessRequest
from ya_market_api.base.sync_config import SyncConfig

from unittest.mock import patch, Mock


class TestSyncOfferAPI:
	def test_get_offer_list_by_business(self):
		session = Mock()
		session.post = Mock()
		session.post.return_value = Mock()
		session.post.return_value.text = "RAW DATA"
		config = SyncConfig(session, "", business_id=1)
		api = SyncOfferAPI(config)
		request = OfferListByBusinessRequest(limit=10, page_token="page-token", category_ids=(1, 2, 3))

		with patch("ya_market_api.offer.sync_api.OfferListByBusinessResponse") as OfferListByBusinessResponseMock:
			OfferListByBusinessResponseMock.model_validate_json = Mock()
			OfferListByBusinessResponseMock.model_validate_json.return_value = "DESERIALIZED DATA"

			with patch.object(api, "validate_response") as validate_response_mock:
				assert api.get_offer_list_by_business() == "DESERIALIZED DATA"
				OfferListByBusinessResponseMock.model_validate_json.assert_called_once_with("RAW DATA")
				validate_response_mock.assert_called_once_with(session.post.return_value)
				session.post.assert_called_once_with(url=api.router.offer_list_by_business(1), params={}, json={})

				session.post.reset_mock()
				assert api.get_offer_list_by_business(request) == "DESERIALIZED DATA"
				session.post.assert_called_once_with(
					url=api.router.offer_list_by_business(1),
					params={"limit": 10, "page_token": "page-token"},
					json={"categoryIds": (1, 2, 3)},
				)



from ya_market_api.const import Header, BASE_URL
from ya_market_api.guide.async_api import AsyncGuideAPI
from ya_market_api.feedback.async_api import AsyncFeedbackAPI
from ya_market_api.offer.async_api import AsyncOfferAPI
from ya_market_api.campaign.async_api import AsyncCampaignAPI
from ya_market_api.order.async_api import AsyncOrderAPI
from ya_market_api.base.async_config import AsyncConfig

from typing import Optional

from aiohttp.client import ClientSession


class AsyncAPI:
	guide: AsyncGuideAPI
	feedback: AsyncFeedbackAPI
	offer: AsyncOfferAPI
	campaign: AsyncCampaignAPI
	order: AsyncOrderAPI
	config: AsyncConfig

	def __init__(self, config: AsyncConfig) -> None:
		self.config = config
		self.guide = AsyncGuideAPI(config)
		self.feedback = AsyncFeedbackAPI(config)
		self.offer = AsyncOfferAPI(config)
		self.campaign = AsyncCampaignAPI(config)
		self.order = AsyncOrderAPI(config)

	async def close(self) -> None:
		await self.config.session.close()

	@classmethod
	async def build(
		cls,
		api_key: str,
		*,
		base_url: str = BASE_URL,
		business_id: Optional[int] = None,
		campaign_id: Optional[int] = None,
		) -> "AsyncAPI":
		config = AsyncConfig(
			await cls.make_session(api_key),
			base_url,
			business_id=business_id,
			campaign_id=campaign_id,
		)

		return cls(config)

	@staticmethod
	async def make_session(api_key: str) -> ClientSession:
		session = ClientSession(headers={Header.API_KEY.value: api_key})
		return session



from ya_market_api.base.sync_api_mixin import SyncAPIMixin
from ya_market_api.order.base_api import BaseOrderAPI
from ya_market_api.order.dataclass import OrderGetRequest, OrderGetResponse


class SyncOrderAPI(SyncAPIMixin, BaseOrderAPI):
	def get_order(self, request: OrderGetRequest) -> OrderGetResponse:
		url = self.router.order_get(self.campaign_id, request.order_id)
		response = self.session.get(url=url)
		self.validate_response(response)
		return OrderGetResponse.model_validate_json(response.text)



from ya_market_api.feedback.dataclass.comment_list import Request

import pytest


class TestRequest:
	def test___init__(self):
		request = Request(comment_ids=(1, 2, 3))
		assert request.comment_ids == (1, 2, 3)

		request = Request(feedback_id=1)
		assert request.feedback_id == 1

	def test_model_dump_request_params(self):
		request = Request(comment_ids=(1, 2, 3))
		assert request.model_dump_request_params() == {}

		request = Request(comment_ids=(1, 2, 3), limit=100, page_token="page-token")
		assert request.model_dump_request_params() == {"limit": 100, "page_token": "page-token"}

	def test_model_dump_request_payload(self):
		request = Request(comment_ids=(1, 2, 3))
		request.feedback_id = 1

		with pytest.warns(UserWarning, match="When using comment_ids, the other parameters will be ignored"):
			assert request.model_dump_request_payload() == {request.S11N_ALIAS_COMMENT_IDS: (1, 2, 3)}

		request.comment_ids = None
		request.feedback_id = None

		with pytest.raises(ValueError, match="feedback_id or comment_ids are required"):
			request.model_dump_request_payload()

		request.comment_ids = (1, 2, 3)
		request.limit = 100
		request.page_token = "page-token"
		assert request.model_dump_request_payload() == {request.S11N_ALIAS_COMMENT_IDS: (1, 2, 3)}



from ya_market_api.exception import InvalidResponseError, AuthorizationError, NotFoundError
from ya_market_api.base.sync_config import SyncConfig

from http import HTTPStatus

from requests.sessions import Session
from requests.models import Response


class SyncAPIMixin:
	config: SyncConfig

	@property
	def session(self) -> Session:
		return self.config.session

	def validate_response(self, response: Response) -> None:
		if not response.ok:
			if response.status_code == HTTPStatus.FORBIDDEN or response.status_code == HTTPStatus.UNAUTHORIZED:
				raise AuthorizationError("Unauthorized")
			elif response.status_code == HTTPStatus.NOT_FOUND:
				raise NotFoundError("Resource was not found")

			raise InvalidResponseError("Response is not valid")



from ya_market_api.base.async_api_mixin import AsyncAPIMixin
from ya_market_api.order.base_api import BaseOrderAPI
from ya_market_api.order.dataclass import OrderGetRequest, OrderGetResponse


class AsyncOrderAPI(AsyncAPIMixin, BaseOrderAPI):
	async def get_order(self, request: OrderGetRequest) -> OrderGetResponse:
		url = self.router.order_get(self.campaign_id, request.order_id)

		async with self.session.get(url=url) as response:
			self.validate_response(response)
			return OrderGetResponse.model_validate_json(await response.text())




class APIError(Exception):
	"""
	Base API error class.
	"""
	pass


class InvalidResponseError(APIError):
	pass


class AuthorizationError(InvalidResponseError):
	pass


class NotFoundError(InvalidResponseError):
	pass


class BusinessIdError(APIError):
	pass


class CampaignIdError(APIError):
	pass



from ya_market_api.base.api import API
from ya_market_api.campaign.router import CampaignRouter


class BaseCampaignAPI(API[CampaignRouter]):
	@staticmethod
	def make_router(base_url: str) -> CampaignRouter:
		return CampaignRouter(base_url)



from ya_market_api.base.api import API
from ya_market_api.exception import BusinessIdError
from ya_market_api.offer.router import OfferRouter


class BaseOfferAPI(API[OfferRouter]):
	@property
	def business_id(self) -> int:
		if self.config.business_id is None:
			raise BusinessIdError("The business_id was not specified")

		return self.config.business_id

	@staticmethod
	def make_router(base_url: str) -> OfferRouter:
		return OfferRouter(base_url)
